import json
import aiobotocore

from asyncio import Lock
from typing import Dict, NamedTuple

from notebook.services.contents.filemanager import FileContentsManager
from tornado import ioloop

from .bookstore_config import BookstoreSettings

from .s3_paths import s3_key, s3_display_path


class ArchiveRecord(NamedTuple):
    filepath: str
    content: str
    queued_time: float


class BookstoreContentsArchiver(FileContentsManager):
    """
    Archives notebooks to S3 on save
    """

    def __init__(self, *args, **kwargs):
        super(FileContentsManager, self).__init__(*args, **kwargs)
        # opt ourselves into being part of the Jupyter App that should have Bookstore Settings applied
        self.settings = BookstoreSettings(parent=self)

        self.log.info("Archiving notebooks to {}".format(s3_display_path(
            self.settings.s3_bucket, self.settings.workspace_prefix)))

        self.session = aiobotocore.get_session()

        # A lock per path to suppress writing to S3 when currently writing
        self.path_locks: Dict[str, Lock] = {}
        # Since locks are not pre-created,
        self.ability_to_create_path_lock = Lock()

    async def archive(self, record: ArchiveRecord):
        """Process a storage write, only allowing one write at a time to each path
        """
        async with self.ability_to_create_path_lock:
            lock = self.path_locks.get(record.filepath)

            if lock is None:
                lock = Lock()
                self.path_locks[record.filepath] = lock

        # Skip writes when a given path is already locked
        if lock.locked():
            self.log.info("Skipping archive of %s", record.filepath)
            return

        async with lock:
            try:
                async with self.session.create_client('s3',
                                                      aws_secret_access_key=self.settings.s3_secret_access_key,
                                                      aws_access_key_id=self.settings.s3_access_key_id,
                                                      endpoint_url=self.settings.s3_endpoint_url,
                                                      region_name=self.settings.s3_region_name,
                                                      ) as client:
                    self.log.info("Processing storage write of %s", record.filepath)
                    file_key = s3_key(self.settings.workspace_prefix, record.filepath)
                    await client.put_object(Bucket=self.settings.s3_bucket, Key=file_key, Body=record.content)
                    self.log.info("Done with storage write of %s", record.filepath)
            except Exception as e:
                    self.log.error(
                        'Error while archiving file: %s %s',
                        record.filepath,
                        e,
                        exc_info=True,
                    )

    def run_pre_save_hook(self, model, path, **kwargs):
        """Store notebook to S3 when saves happen
        """
        if model["type"] != "notebook":
            return

        content = json.dumps(model["content"])

        loop = ioloop.IOLoop.current()

        # Offload the archival to be scheduled to work on the event loop
        loop.spawn_callback(
            self.archive,
            ArchiveRecord(
                content=content,
                filepath=path,
                queued_time=ioloop.IOLoop.current().time(),
            ),
        )



from traitlets import Int, Unicode
from traitlets.config import LoggingConfigurable


class BookstoreSettings(LoggingConfigurable):
    """The same settings to be shared across archival, publishing, and scheduling
    """

    workspace_prefix = Unicode(
        "workspace", help="Prefix for the live workspace notebooks"
    ).tag(config=True)
    published_prefix = Unicode("published", help="Prefix for published notebooks").tag(
        config=True
    )

    ## S3 Settings for the S3 backed storage (other implementations can add on below)
    # Allowed to not set these as we can pick up IAM roles instead
    s3_access_key_id = Unicode(
        help="S3/AWS access key ID", allow_none=True, default_value=None
    ).tag(config=True, env="JPYNB_S3_ACCESS_KEY_ID")
    s3_secret_access_key = Unicode(
        help="S3/AWS secret access key", allow_none=True, default_value=None
    ).tag(config=True, env="JPYNB_S3_SECRET_ACCESS_KEY")

    s3_endpoint_url = Unicode("https://s3.amazonaws.com", help="S3 endpoint URL").tag(
        config=True, env="JPYNB_S3_ENDPOINT_URL"
    )
    s3_region_name = Unicode("us-east-1", help="Region name").tag(
        config=True, env="JPYNB_S3_REGION_NAME"
    )
    s3_bucket = Unicode("bookstore", help="Bucket name to store notebooks").tag(
        config=True, env="JPYNB_S3_BUCKET"
    )

    max_threads = Int(
        16,
        help="Maximum number of threads for the threadpool allocated for S3 read/writes",
    ).tag(config=True)



# Our S3 path delimiter will remain fixed as '/' in all uses
delimiter = "/"


def _join(*args):
    '''Join S3 bucket args together, removing empty entries and stripping left-leading
    '''
    return delimiter.join(filter(lambda s: s != '', map(lambda s: s.lstrip(delimiter), args)))


def s3_path(bucket, prefix, path=''):
    """compute the s3 path based on the bucket, prefix, and the path to the notebook"""
    return _join(bucket, prefix, path)


def s3_key(prefix, path=''):
    """compute the s3 key based on the prefix, and the path to the notebook"""
    return _join(prefix, path)


def s3_display_path(bucket, prefix, path=''):
    """create a display name for use in logs"""
    return 's3://' + s3_path(bucket, prefix, path)



import json

import aiobotocore

from notebook.base.handlers import APIHandler, path_regex
from notebook.utils import url_path_join
from tornado import web

from ._version import __version__, version_info
from .bookstore_config import BookstoreSettings
from .s3_paths import s3_path, s3_key, s3_display_path


version = __version__


class BookstoreVersionHandler(APIHandler):
    """Returns the version of bookstore currently running. Used mostly to lay foundations
    for this package though frontends can use this endpoint for feature detection.
    """

    @web.authenticated
    def get(self):
        self.finish(json.dumps({"bookstore": True, "version": version}))


# NOTE: We need to ensure that publishing is not configured if bookstore settings are not
# set. Because of how the APIHandlers cannot be configurable, all we can do is reach into settings
# For applications this will mean checking the config and then applying it in


class BookstorePublishHandler(APIHandler):
    """Publish a notebook to the publish path"""

    def __init__(self, *args, **kwargs):
        super(APIHandler, self).__init__(*args, **kwargs)
        # create an easy helper to get at our bookstore settings quickly
        self.bookstore_settings = BookstoreSettings(config=self.config)

        self.session = aiobotocore.get_session()

    async def _publish(self, model, path):
        if model['type'] != 'notebook':
            raise web.HTTPError(400, "bookstore only publishes notebooks")
        content = model['content']

        full_s3_path = s3_path(
            self.bookstore_settings.s3_bucket, self.bookstore_settings.published_prefix, path
        )
        file_key = s3_key(self.bookstore_settings.published_prefix, path)

        self.log.info(
            "Publishing to %s",
            s3_display_path(
                self.bookstore_settings.s3_bucket, self.bookstore_settings.published_prefix, path
            ),
        )
        async with self.session.create_client(
            's3',
            aws_secret_access_key=self.bookstore_settings.s3_secret_access_key,
            aws_access_key_id=self.bookstore_settings.s3_access_key_id,
            endpoint_url=self.bookstore_settings.s3_endpoint_url,
            region_name=self.bookstore_settings.s3_region_name,
        ) as client:
            self.log.info("Processing published write of %s", path)
            obj = await client.put_object(
                Bucket=self.bookstore_settings.s3_bucket, Key=file_key, Body=json.dumps(content)
            )
            self.log.info("Done with published write of %s", path)

        self.set_status(201)

        resp_content = {"s3path": full_s3_path}

        if 'VersionId' in obj:
            resp_content["versionID"] = obj['VersionId']

        resp_str = json.dumps(resp_content)
        self.finish(resp_str)

    @web.authenticated
    async def put(self, path=''):
        """Publish a notebook on a given path. The payload for this directly matches that of the contents API for PUT.
        """
        self.log.info("About to publish %s", path)

        if path == '' or path == '/':
            raise web.HTTPError(400, "Must have a path to publish to")

        model = self.get_json_body()

        if model:
            await self._publish(model, path.lstrip('/'))
        else:
            raise web.HTTPError(400, "Cannot publish empty model")


def load_jupyter_server_extension(nb_app):
    web_app = nb_app.web_app
    host_pattern = '.*$'
    base_bookstore_pattern = url_path_join(web_app.settings['base_url'], '/api/bookstore')

    # Always enable the version handler
    web_app.add_handlers(host_pattern, [(base_bookstore_pattern, BookstoreVersionHandler)])

    if not nb_app.config.get("BookstoreSettings"):
        nb_app.log.info("Not enabling bookstore publishing since bookstore endpoint not configured")
    else:
        web_app.add_handlers(
            host_pattern,
            [
                (
                    url_path_join(base_bookstore_pattern, r"/published%s" % path_regex),
                    BookstorePublishHandler,
                )
            ],
        )



from .archive import BookstoreContentsArchiver
from .bookstore_config import BookstoreSettings
from .handlers import load_jupyter_server_extension
from ._version import version_info, __version__

def _jupyter_server_extension_paths():
    return [dict(module="bookstore")]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""""setup.py

Note: Do a Python check

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
from __future__ import print_function

from io import open
import os
from os import path
import sys

from setuptools import setup


v = sys.version_info

if v[:2] < (3, 6):
    print('ERROR: Bookstore requires Python 3.6 or higher', file=sys.stderr)
    sys.exit(1)

# We have the correct Python version, proceed.

pjoin = os.path.join

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(pjoin(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the bookstore version
ns = {}
with open(pjoin(here, 'bookstore', '_version.py')) as f:
    exec(f.read(), {}, ns)


target_dir = pjoin("etc", "jupyter", "jupyter_notebook_config.d")
config_files = [pjoin("jupyter_config", "jupyter_notebook_config.d", "bookstore.json")]
data_files = [(target_dir, config_files)]

setup(
    name='bookstore',
    version=ns['__version__'],
    description='Storage Workflows for Notebooks',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nteract/bookstore',
    author='nteract contributors',
    author_email='nteract@googlegroups.com',
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='jupyter storage nteract notebook',
    packages=['bookstore'],
    install_requires=[
        'future',
        'futures ; python_version < "3.0"',
        'ipython >= 5.0',
        'notebook',
        's3fs',
        'tornado >= 5.1.1',
        'aiobotocore==0.10.0',
    ],
    extras_require={
        'test': [
            'codecov',
            'coverage',
            'mock',
            'mypy==0.660',
            'pytest>=3.3',
            'pytest-cov',
            'pytest-mock',
            'black',
        ],
    },
    data_files=data_files,
    entry_points={},
    project_urls={
        'Documentation': 'https://github.com/nteract/bookstore/#todo',
        'Funding': 'https://nteract.io',
        'Source': 'https://github.com/nteract/bookstore/',
        'Tracker': 'https://github.com/nteract/bookstore/issues',
    },

)



"""Tests for s3 paths"""
import pytest

from bookstore.s3_paths import s3_display_path


def test_s3_paths():
    bucket = 'mybucket'
    prefix = 'yo'
    path = 'pickles'
    assert s3_display_path(bucket, prefix, path) == 's3://mybucket/yo/pickles'


def test_s3_paths_no_path():
    bucket = 'mybucket'
    prefix = 'yo'
    assert s3_display_path(bucket, prefix) == 's3://mybucket/yo'


def test_s3_paths_no_prefix():
    bucket = 'mybucket'
    path = 'pickles'
    with pytest.raises(NameError):
        s3_display_path(bucket, prefix, path)


def test_join():
    pass
